import base64
import typing as t

from pydantic import PlainSerializer, PlainValidator, WithJsonSchema


def serialize_base64bytes(value: bytes) -> str:
    return base64.b64encode(value).decode("ascii")


def validate_base64bytes(value: t.Any) -> bytes:
    match value:
        case bytes():
            return value
        case str():
            return base64.b64decode(value)
        case _:
            raise ValueError(f"Invalid input type: {type(value)}")


Base64Bytes = t.Annotated[
    bytes,
    PlainValidator(validate_base64bytes),
    PlainSerializer(serialize_base64bytes),
    WithJsonSchema({"type": "string", "format": "base64"}),
]



from typing import Optional

from pydantic import BaseModel, Field

from . import ToolOutput


class MergeContentInput(BaseModel):
    """Input data model for MergeContent node."""

    separator: str = Field(
        default="\n\n\n\n", description="Separator to use between merged content sections"
    )
    content_headers: list[str] = Field(
        default=[], description="Headers for content sections"
    )
    content_texts: list[str] = Field(default=[], description="Texts for content sections")
    indent_level: int = Field(
        default=0, description="Indent level for the merged content"
    )
    title: Optional[str] = Field(default=None, description="Title for the merged content")


class MergeContentOutput(ToolOutput):
    """Output data model for MergeContent node."""

    merged_content: str = Field(
        default="",
        description="Merged content from multiple nodes with formatted headers",
    )



from typing import Any

from ..client import A79Client
from ..models.tools import DEFAULT
from ..models.tools.thoughtspot_models import (
    ConnectionDetails,
    ConnectionResponse,
    CreateConnectionInput,
    CreateConnectionOutput,
    CreateLiveboardInput,
    CreateLiveboardOutput,
    PinAnswerInput,
    PinAnswerOutput,
    SearchDataInput,
    SearchDataOutput,
    SearchDataResponse,
    SearchMetadataInput,
    SearchMetadataOutput,
    TableInfo,
)

__all__ = [
    "ConnectionDetails",
    "ConnectionResponse",
    "CreateConnectionInput",
    "CreateConnectionOutput",
    "CreateLiveboardInput",
    "CreateLiveboardOutput",
    "PinAnswerInput",
    "PinAnswerOutput",
    "SearchDataInput",
    "SearchDataOutput",
    "SearchDataResponse",
    "SearchMetadataInput",
    "SearchMetadataOutput",
    "TableInfo",
    "create_connection",
    "create_liveboard",
    "pin_answer",
    "search_metadata",
    "search_data",
]


def create_connection(
    *,
    data_store_connector_id: str | None = DEFAULT,
    data_store_connector_name: str | None = DEFAULT,
    data_store_config: dict[str, Any] = DEFAULT,
) -> CreateConnectionOutput:
    """
    "Creates a connection to a data source in Thoughtspot."
    """
    kwargs = locals()
    kwargs = {k: v for k, v in kwargs.items() if v is not DEFAULT}
    input_model = CreateConnectionInput.model_validate(kwargs)

    client = A79Client()
    output_model = client.execute_tool(
        package="thoughtspot", name="create_connection", input=input_model.model_dump()
    )
    return CreateConnectionOutput.model_validate(output_model)


def create_liveboard(
    *, liveboard_name: str, liveboard_description: str | None = DEFAULT
) -> CreateLiveboardOutput:
    """
    Creates a new liveboard in Thoughtspot to host visualizations.
    """
    kwargs = locals()
    kwargs = {k: v for k, v in kwargs.items() if v is not DEFAULT}
    input_model = CreateLiveboardInput.model_validate(kwargs)

    client = A79Client()
    output_model = client.execute_tool(
        package="thoughtspot", name="create_liveboard", input=input_model.model_dump()
    )
    return CreateLiveboardOutput.model_validate(output_model)


def pin_answer(
    *,
    table_id: str,
    answer_name: str,
    liveboard_id: str,
    chart_type: str,
    answer_data: SearchDataResponse,
) -> PinAnswerOutput:
    """
    "Pins an answer to a liveboard in Thoughtspot.
    """
    kwargs = locals()
    kwargs = {k: v for k, v in kwargs.items() if v is not DEFAULT}
    input_model = PinAnswerInput.model_validate(kwargs)

    client = A79Client()
    output_model = client.execute_tool(
        package="thoughtspot", name="pin_answer", input=input_model.model_dump()
    )
    return PinAnswerOutput.model_validate(output_model)


def search_metadata(*, query: str) -> SearchMetadataOutput:
    """
    Searches metadata in Thoughtspot.
    """
    kwargs = locals()
    kwargs = {k: v for k, v in kwargs.items() if v is not DEFAULT}
    input_model = SearchMetadataInput.model_validate(kwargs)

    client = A79Client()
    output_model = client.execute_tool(
        package="thoughtspot", name="search_metadata", input=input_model.model_dump()
    )
    return SearchMetadataOutput.model_validate(output_model)


def search_data(*, query: str, logical_table_identifier: str) -> SearchDataOutput:
    """
    Searches data in Thoughtspot.
    """
    kwargs = locals()
    kwargs = {k: v for k, v in kwargs.items() if v is not DEFAULT}
    input_model = SearchDataInput.model_validate(kwargs)

    client = A79Client()
    output_model = client.execute_tool(
        package="thoughtspot", name="search_data", input=input_model.model_dump()
    )
    return SearchDataOutput.model_validate(output_model)



import json
import typing as t

import pandas as pd
from pydantic import BaseModel, Field

from . import citation_models

DEFAULT: t.Any = object()
"""
Sentinel value for when a field has not been provided as input.
"""


class HumanReadableNodeOutput(BaseModel):
    """
    Has presentation details of how to represent or expose node-output to the caller.
    For example, if a node generates an table in GCS, the NodeOutputSummary could be
    returning a preview of few rows of the table.
    """

    text: str = Field(default="")
    citations: list[citation_models.Citation] | None = Field(default_factory=lambda: [])
    table_data: dict[str, t.Any] | None = Field(default_factory=lambda: None)
    datasource: int | None = Field(
        default=None, description="Reference to a datasource by its ID"
    )
    liveboard: str | None = Field(
        default=None, description="Reference to a liveboard by its ID"
    )
    stream_id: str | None = Field(
        default=None, description="Reference to a stream by its ID"
    )

    @staticmethod
    def convert_to_table_data(df: pd.DataFrame) -> dict:
        """Convert a DataFrame to our standard table data format.

        Needs to be json serializable as this is stored in the database.

        Args:
            df: pandas DataFrame to convert

        Returns:
            dict: Table data in standard format with columns and data
        """
        # Convert dates to ISO format to make it json serializable.
        records = json.loads(df.to_json(orient="records", date_format="iso"))

        table_data = {
            "columns": [
                {"name": str(col), "type": str(df[col].dtype)} for col in df.columns
            ],
            "data": records,
        }
        return table_data

    @staticmethod
    def represents_table_data(var: t.Any) -> bool:
        """Check if a variable can be converted to a pandas DataFrame.

        Args:
            var: Variable to check

        Returns:
            bool: True if variable can be converted to DataFrame, False otherwise
        """
        # Check that the variable is a list
        if not isinstance(var, list):
            return False

        if len(var) == 0:
            return False

        # Check that each element in the list is a dict
        if not all(isinstance(item, dict) for item in var):
            return False

        # Check that all dictionaries have the same keys
        first_keys = set(var[0].keys())
        if not all(set(item.keys()) == first_keys for item in var):
            return False

        try:
            # Try converting directly to DataFrame
            records = pd.DataFrame.from_records(var)
            return len(records) > 0
        except (ValueError, TypeError):
            pass
        return False


class ToolSummary(BaseModel):
    short_summary: str
    long_summary: HumanReadableNodeOutput = Field(default_factory=HumanReadableNodeOutput)


class ToolOutput(BaseModel):
    tool_summary: ToolSummary



from ..client import A79Client
from ..models.tools import DEFAULT
from ..models.tools.file_models import (
    GetFileContentInput,
    GetFileContentOutput,
    GetFileSearchInput,
)
from ..models.tools.folder_models import SearchOutput

__all__ = [
    "GetFileContentInput",
    "GetFileContentOutput",
    "GetFileSearchInput",
    "SearchOutput",
    "read_csv",
    "read_pdf",
    "read_text",
    "extract_audio_transcript",
    "search",
]


def read_csv(*, file_name: str) -> GetFileContentOutput:
    """Read CSV file content and return as DataFrame.

    This tool specifically handles CSV files and returns the content as a
    structured DataFrame along with metadata.

    Args:
        input: GetFileContentInput containing file_name to query

    Returns:
        GetFileContentOutput containing CSV data as DataFrame and metadata
    """
    kwargs = locals()
    kwargs = {k: v for k, v in kwargs.items() if v is not DEFAULT}
    input_model = GetFileContentInput.model_validate(kwargs)

    client = A79Client()
    output_model = client.execute_tool(
        package="file", name="read_csv", input=input_model.model_dump()
    )
    return GetFileContentOutput.model_validate(output_model)


def read_pdf(*, file_name: str) -> GetFileContentOutput:
    """Read PDF file content and return as extracted text.

    This tool specifically handles PDF files and returns the extracted text
    content along with PDF-specific metadata.

    Args:
        input: GetFileContentInput containing file_name to query

    Returns:
        GetFileContentOutput containing PDF text content and metadata
    """
    kwargs = locals()
    kwargs = {k: v for k, v in kwargs.items() if v is not DEFAULT}
    input_model = GetFileContentInput.model_validate(kwargs)

    client = A79Client()
    output_model = client.execute_tool(
        package="file", name="read_pdf", input=input_model.model_dump()
    )
    return GetFileContentOutput.model_validate(output_model)


def read_text(*, file_name: str) -> GetFileContentOutput:
    """Read plain text file content.

    This tool specifically handles plain text files and returns the raw
    text content.

    Args:
        input: GetFileContentInput containing file_name to query

    Returns:
        GetFileContentOutput containing plain text content and metadata
    """
    kwargs = locals()
    kwargs = {k: v for k, v in kwargs.items() if v is not DEFAULT}
    input_model = GetFileContentInput.model_validate(kwargs)

    client = A79Client()
    output_model = client.execute_tool(
        package="file", name="read_text", input=input_model.model_dump()
    )
    return GetFileContentOutput.model_validate(output_model)


def extract_audio_transcript(*, file_name: str) -> GetFileContentOutput:
    """Extract transcript from audio file.

    This tool specifically handles audio files and returns the extracted
    transcript text.

    Args:
        input: GetFileContentInput containing file_name to query

    Returns:
        GetFileContentOutput containing audio transcript and metadata
    """
    kwargs = locals()
    kwargs = {k: v for k, v in kwargs.items() if v is not DEFAULT}
    input_model = GetFileContentInput.model_validate(kwargs)

    client = A79Client()
    output_model = client.execute_tool(
        package="file", name="extract_audio_transcript", input=input_model.model_dump()
    )
    return GetFileContentOutput.model_validate(output_model)


def search(*, file_name: str, query: str, num_results: int = DEFAULT) -> SearchOutput:
    """Search for content within a specific file using semantic and lexical search.

    This tool performs hybrid search (combining semantic and lexical search) within
    a specific file using Vespa indexing. Results are ranked by relevance score
    and include highlighted snippets.

    Args:
        input: SearchInput containing folder_path (treated as file_name) and search query

    Returns:
        SearchOutput containing ranked search results with scores and snippets
    """
    kwargs = locals()
    kwargs = {k: v for k, v in kwargs.items() if v is not DEFAULT}
    input_model = GetFileSearchInput.model_validate(kwargs)

    client = A79Client()
    output_model = client.execute_tool(
        package="file", name="search", input=input_model.model_dump()
    )
    return SearchOutput.model_validate(output_model)



import typing as t

from pydantic import BaseModel, Field

from . import ToolOutput


class GetCompanyInfoInput(BaseModel):
    domain: str
    data: dict[str, t.Any] = Field(
        default_factory=dict, description="Parameter to provide payload."
    )
    timeout: float = Field(default=60, description="Request timeout in seconds")


class GetCompanyInfoOutput(ToolOutput):
    content: dict[str, t.Any]


class SearchCompaniesInput(BaseModel):
    data: dict[str, t.Any] = Field(
        default_factory=dict, description="Parameter to provide payload."
    )
    timeout: float = Field(default=60, description="Request timeout in seconds")


class SearchCompaniesOutput(ToolOutput):
    content: dict[str, t.Any]


class GetMarketIntelligenceInput(BaseModel):
    company_identifier: str = Field(
        default="", description="Parameter to provide payload."
    )
    timeout: float = Field(default=60, description="Request timeout in seconds")


class GetMarketIntelligenceOutput(ToolOutput):
    content: dict[str, t.Any]


class GetCompanyByCompanyIdentifierInput(BaseModel):
    company_identifier: str = Field(
        default="", description="Parameter to provide payload."
    )
    timeout: float = Field(default=60, description="Request timeout in seconds")


class GetCompanyByCompanyIdentifierOutput(ToolOutput):
    content: dict[str, t.Any]


class GetSourcesByCompanyIdentifierInput(BaseModel):
    company_identifier: str = Field(
        default="", description="Parameter to provide payload."
    )
    timeout: float = Field(default=60, description="Request timeout in seconds")


class GetSourcesByCompanyIdentifierOutput(ToolOutput):
    content: dict[str, t.Any]


class SearchSourcesInput(BaseModel):
    data: dict[str, t.Any] = Field(
        default_factory=dict, description="Parameter to provide payload."
    )
    timeout: float = Field(default=60, description="Request timeout in seconds")


class SearchSourcesOutput(ToolOutput):
    content: dict[str, t.Any]



from typing import Any

from pydantic import BaseModel

from ..client import A79Client
from ..models.tools import DEFAULT
from ..models.tools.llm_models import (
    ChatInput,
    ChatOutput,
    Enum,
    ErrorHandling,
    InferenceMode,
    LLMUsageData,
)
from ..models.tools.workflow_models import ToolStreamResponse

__all__ = [
    "ChatInput",
    "ChatOutput",
    "Enum",
    "ErrorHandling",
    "InferenceMode",
    "LLMUsageData",
    "chat",
    "chat_stream",
]


def chat(
    *,
    schema: dict[str, Any] | type[BaseModel] | None = DEFAULT,
    model: str = DEFAULT,
    prompt_str: str | None = DEFAULT,
    temperature: float = DEFAULT,
    stop: list[str] | None = DEFAULT,
    error_handling: ErrorHandling = DEFAULT,
    top_p: float | None = DEFAULT,
    seed: int | None = DEFAULT,
    presence_penalty: float | None = DEFAULT,
    frequency_penalty: float | None = DEFAULT,
    tool_choice: str | None = DEFAULT,
    inference_mode: InferenceMode = DEFAULT,
    output_schema: dict[str, Any] | type[BaseModel] | None = DEFAULT,
    stream: bool = DEFAULT,
) -> ChatOutput:
    """
    AI powered chat
    """
    kwargs = locals()
    kwargs = {k: v for k, v in kwargs.items() if v is not DEFAULT}
    input_model = ChatInput.model_validate(kwargs)

    client = A79Client()
    output_model = client.execute_tool(
        package="llm", name="chat", input=input_model.model_dump()
    )
    return ChatOutput.model_validate(output_model)


def chat_stream(
    *,
    schema: dict[str, Any] | type[BaseModel] | None = DEFAULT,
    model: str = DEFAULT,
    prompt_str: str | None = DEFAULT,
    temperature: float = DEFAULT,
    stop: list[str] | None = DEFAULT,
    error_handling: ErrorHandling = DEFAULT,
    top_p: float | None = DEFAULT,
    seed: int | None = DEFAULT,
    presence_penalty: float | None = DEFAULT,
    frequency_penalty: float | None = DEFAULT,
    tool_choice: str | None = DEFAULT,
    inference_mode: InferenceMode = DEFAULT,
    output_schema: dict[str, Any] | type[BaseModel] | None = DEFAULT,
    stream: bool = DEFAULT,
) -> ToolStreamResponse:
    """
    A streaming version of the chat tool that returns a StreamingResponse.

    This function is designed for use with server-sent events (SSE) and frameworks
    like FastAPI. The returned StreamingResponse can be used directly in web endpoints.

    Args:
        input: ChatInput containing the prompt and options

    Returns:
        StreamingResponse: StreamingResponse object containing the SSE data chunks

    Example:
        ```python
        # Use directly in FastAPI endpoint
        @app.get("/stream")
        def stream_endpoint():
            return llm.chat_stream(chat_input)

        # Or return the response directly
        streaming_response = llm.chat_stream(chat_input)
        ```
    """
    kwargs = locals()
    kwargs = {k: v for k, v in kwargs.items() if v is not DEFAULT}
    input_model = ChatInput.model_validate(kwargs)

    client = A79Client()
    stream_id = client.execute_tool_streaming(
        package="llm", name="chat_stream", input=input_model.model_dump()
    )
    return ToolStreamResponse(
        stream_id=stream_id, tool_name="chat_stream", package_name="llm"
    )



import enum
from typing import Any

from pydantic import BaseModel, Field

from ..models.tools import ToolOutput


class InputMethod(str, enum.Enum):
    console = "console"
    api = "api"


class InputAction(str, enum.Enum):
    text = "text"
    row_selection = "row_selection"
    row_selection_with_feedback = "row_selection_with_feedback"
    chart_config_selection = "chart_config_selection"
    worksheet_document_upload = "worksheet_document_upload"


class HumanFeedbackToolInput(BaseModel):
    input: str = Field(default="")
    tabular_data: list[dict[str, Any]] = Field(default_factory=list)
    chart_config_data: dict[str, Any] | None = None
    input_method: InputMethod = InputMethod.console
    input_action: InputAction = Field(default=InputAction.text)
    input_question: str = Field(default="")
    worksheet_id: int | None = Field(default=None)


class HumanFeedbackToolOutput(ToolOutput):
    content: dict[str, Any] = Field(default_factory=dict)
    input: str = Field(default="")
    tabular_data: dict[str, Any] = Field(default_factory=dict)
    chart_config_data: dict[str, Any] | None = None
    input_action: InputAction = Field(default=InputAction.text)
    input_question: str = Field(default="")
    worksheet_id: int | None = Field(default=None)



from pydantic import BaseModel

from . import ToolOutput


class GetLatestCompanyFilingTextInput(BaseModel):
    company_identifier: str
    filing_types: list[str]


class GetLatestCompanyFilingTextOutput(ToolOutput):
    content: dict[str, str]



from pydantic import BaseModel


class ErrorHandling(BaseModel):
    """
    Configuration for error handling in nodes.

    Attributes:
        timeout_seconds (float | None): Timeout in seconds for node execution.
        retry_interval_seconds (float): Interval between retries in seconds.
        max_retries (int): Maximum number of retries.
        backoff_rate (float): Rate of increase for retry intervals.
    """

    timeout_seconds: float | None = (
        900  # 15 minutes because sub-workflows can be run in the python node
    )
    retry_interval_seconds: float = 1
    max_retries: int = 0
    backoff_rate: float = 1



# mypy: ignore-errors
# This file has a lot of type based creation of models, which mypy is flagging as errors.
# Turning off type checking for this file.

import inspect
import sys
from typing import (
    Annotated,
    Any,
    Callable,
    Final,
    ForwardRef,
    Literal,
    NewType,
    Optional,
    TypedDict,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from pydantic import BaseModel, Field, create_model


def is_pydantic_model(obj: Any) -> bool:
    return inspect.isclass(obj) and issubclass(obj, BaseModel)


def serialize_type(typ):
    """Serialize a type annotation to a structured representation"""
    if isinstance(typ, type) and issubclass(typ, BaseModel):
        # For Pydantic models, reference by name
        return {"kind": "model", "name": typ.__name__}

    elif get_origin(typ) is not None:
        # Handle generic types like list, dict, etc.
        origin = get_origin(typ)
        args = get_args(typ)

        origin_name = origin.__name__ if hasattr(origin, "__name__") else str(origin)

        if origin in (list, list[Any]):
            return {"kind": "list", "item_type": serialize_type(args[0])}
        elif origin in (dict, dict[Any, Any]):
            return {
                "kind": "dict",
                "key_type": serialize_type(args[0]),
                "value_type": serialize_type(args[1]),
            }
        elif origin in (set, set[Any]):
            return {"kind": "set", "item_type": serialize_type(args[0])}
        elif origin in (tuple, tuple[Any, ...]):
            return {"kind": "tuple", "item_types": [serialize_type(arg) for arg in args]}
        elif origin in (Union, Union[Any, ...]):
            # Special handling for Optional[T] which is Union[T, None]
            if len(args) == 2 and args[1] is type(None):
                return {"kind": "optional", "type": serialize_type(args[0])}
            return {"kind": "union", "types": [serialize_type(arg) for arg in args]}
        elif origin is Literal:
            return {"kind": "literal", "values": args}
        elif origin is Annotated:
            return {
                "kind": "annotated",
                "type": serialize_type(args[0]),
                "metadata": [str(arg) for arg in args[1:]],
            }
        elif origin is Callable:
            if len(args) == 2 and args[0] is Ellipsis:
                return {"kind": "callable", "return_type": serialize_type(args[1])}
            return {
                "kind": "callable",
                "arg_types": [serialize_type(arg) for arg in args[:-1]],
                "return_type": serialize_type(args[-1]),
            }
        elif origin is type:
            return {"kind": "type", "type": serialize_type(args[0])}
        elif origin is Final:
            return {"kind": "final", "type": serialize_type(args[0])}
        else:
            return {
                "kind": "generic",
                "origin": origin_name,
                "args": [serialize_type(arg) for arg in args],
            }

    elif isinstance(typ, ForwardRef):
        # Handle forward references
        return {"kind": "forward_ref", "name": typ.__forward_arg__}

    elif isinstance(typ, TypeVar):
        # Handle type variables
        return {
            "kind": "typevar",
            "name": typ.__name__,
            "constraints": [serialize_type(c) for c in typ.__constraints__]
            if typ.__constraints__
            else None,
            "bound": serialize_type(typ.__bound__) if typ.__bound__ is not None else None,
        }
    elif hasattr(typ, "__supertype__"):  # Handle NewType
        return {
            "kind": "newtype",
            "name": typ.__name__,
            "supertype": serialize_type(typ.__supertype__),
        }

    else:
        # Handle simple types
        try:
            return {
                "kind": "simple",
                "name": typ.__name__ if hasattr(typ, "__name__") else str(typ),
            }
        except AttributeError:
            return {"kind": "simple", "name": str(typ)}


def deserialize_type(
    type_info: dict[str, Any], model_registry: Optional[dict[str, Any]] = None
) -> Any:
    """Deserialize a type from its structured representation"""
    if model_registry is None:
        model_registry = {}

    kind = type_info.get("kind")

    if kind == "model":
        # Handle Pydantic model references
        model_name = type_info["name"]
        if model_name in model_registry:
            return model_registry[model_name]
        else:
            # Return a ForwardRef if model not yet defined
            return ForwardRef(model_name)

    elif kind == "list":
        # Handle List types
        item_type = deserialize_type(type_info["item_type"], model_registry)
        return list[item_type]

    elif kind == "dict":
        # Handle Dict types
        key_type = deserialize_type(type_info["key_type"], model_registry)
        value_type = deserialize_type(type_info["value_type"], model_registry)
        return dict[key_type, value_type]

    elif kind == "set":
        # Handle Set types
        item_type = deserialize_type(type_info["item_type"], model_registry)
        return set[item_type]

    elif kind == "tuple":
        # Handle Tuple types
        item_types = [
            deserialize_type(t, model_registry) for t in type_info["item_types"]
        ]
        return tuple[tuple(item_types)]

    elif kind == "optional":
        # Handle Optional types
        inner_type = deserialize_type(type_info["type"], model_registry)
        return Optional[inner_type]

    elif kind == "union":
        # Handle Union types
        types = [deserialize_type(t, model_registry) for t in type_info["types"]]
        return Union[tuple(types)]
    elif kind == "type":
        # Handle Type types
        inner_type = deserialize_type(type_info["type"], model_registry)
        return type[inner_type]

    elif kind == "final":
        # Handle Final types
        inner_type = deserialize_type(type_info["type"], model_registry)
        return Final[inner_type]

    elif kind == "typevar":
        # Handle TypeVar
        constraints = None
        if type_info["constraints"]:
            constraints = tuple(
                deserialize_type(c, model_registry) for c in type_info["constraints"]
            )
        bound = None
        if type_info["bound"]:
            bound = deserialize_type(type_info["bound"], model_registry)
        return TypeVar(type_info["name"], constraints=constraints, bound=bound)

    elif kind == "protocol":
        # Handle Protocol
        return ForwardRef(type_info["name"])

    elif kind == "typeddict":
        # Handle TypedDict
        fields = {
            k: deserialize_type(v, model_registry) for k, v in type_info["fields"].items()
        }
        return TypedDict(type_info["name"], fields)

    elif kind == "newtype":
        # Handle NewType
        supertype = deserialize_type(type_info["supertype"], model_registry)
        return NewType(type_info["name"], supertype)

    elif kind == "forward_ref":
        # Handle forward references
        return ForwardRef(type_info["name"])

    elif kind == "simple":
        # Handle simple types
        type_name = type_info["name"]
        if type_name == "str":
            return str
        elif type_name == "int":
            return int
        elif type_name == "float":
            return float
        elif type_name == "bool":
            return bool
        elif type_name == "None" or type_name == "NoneType":
            return type(None)
        else:
            # Try to resolve from built-in types
            if hasattr(sys.modules["builtins"], type_name):
                return getattr(sys.modules["builtins"], type_name)
            # Default to Any if type can't be resolved
            return Any

    else:
        # Default case
        return Any


def serialize_model_definition(model_class: type[BaseModel]) -> dict[str, Any]:
    """Serialize a Pydantic model class definition to a dictionary."""
    # Get field information including types
    fields = {}
    type_hints = get_type_hints(model_class)

    for field_name, field_info in model_class.model_fields.items():
        field_type = type_hints.get(field_name, Any)

        # Serialize field information
        field_data = {
            "type": serialize_type(field_type),
            "required": field_info.is_required(),
            "default": None,  # Will be updated if not required
            "description": field_info.description,
        }

        # Handle default values
        if not field_info.is_required():
            default_value = field_info.get_default()
            if default_value is not None and hasattr(default_value, "model_dump"):
                # Handle default value that is a Pydantic model
                field_data["default"] = {
                    "is_model": True,
                    "value": default_value.model_dump(),
                }
            else:
                # Simple default value
                field_data["default"] = default_value

        fields[field_name] = field_data

    # Include any validators and extra model info
    model_config = {}
    if hasattr(model_class, "model_config"):
        for key, value in model_class.model_config.items():
            if isinstance(value, (str, int, float, bool, list, dict)) or value is None:
                model_config[key] = value

    # Generate a basic schema without using model_json_schema
    schema = {"type": "object", "properties": {}, "required": []}

    for field_name, field_info in fields.items():
        # Add to properties
        schema["properties"][field_name] = {
            "type": "object" if field_info["type"]["kind"] == "model" else "string",
            "description": field_info["description"] or "",
        }

        # Add to required if needed
        if field_info["required"]:
            schema["required"].append(field_name)

    return {
        "name": model_class.__name__,
        "fields": fields,
        "schema": schema,
        "config": model_config,
    }


def collect_model_dependencies(
    model_class: type[BaseModel], collected_models=None, localns=None
):
    """Recursively collect all model dependencies."""
    if collected_models is None:
        collected_models = {}

    # Skip if already collected
    if model_class.__name__ in collected_models:
        return collected_models

    # Add current model
    collected_models[model_class.__name__] = model_class

    # Get the module's globals and locals for type resolution
    if localns is None:
        localns = sys.modules[model_class.__module__].__dict__

    # Find all field types that are Pydantic models
    type_hints = get_type_hints(model_class, localns=localns)

    for field_type in type_hints.values():
        # Process the type to find nested models
        def process_type(typ):
            if isinstance(typ, type) and issubclass(typ, BaseModel):
                collect_model_dependencies(typ, collected_models, localns)
            elif get_origin(typ) is not None:
                # Handle container types
                for arg in get_args(typ):
                    process_type(arg)
            elif isinstance(typ, ForwardRef):
                # Try to resolve the forward reference
                try:
                    resolved_type = typ._evaluate(localns, None, set())
                    if isinstance(resolved_type, type) and issubclass(
                        resolved_type, BaseModel
                    ):
                        collect_model_dependencies(
                            resolved_type, collected_models, localns
                        )
                except (NameError, AttributeError):
                    # If we can't resolve it yet, that's okay
                    # - it will be handled during deserialization
                    pass

        process_type(field_type)

    return collected_models


def serialize_all_models(root_model: type[BaseModel]) -> dict[str, Any]:
    """Serialize a model and all its dependencies into a single dictionary"""
    # First collect all models
    all_models = collect_model_dependencies(root_model)

    # Then serialize each model
    serialized_models = {}
    for model_name, model_class in all_models.items():
        serialized_models[model_name] = serialize_model_definition(model_class)

    return {"models": serialized_models, "root_model": root_model.__name__}


def get_model_dependencies(model_def: dict[str, Any]) -> set[str]:
    """Get all model dependencies for a given model definition."""
    dependencies = set()

    def process_type_info(type_info: dict[str, Any]):
        if type_info["kind"] == "model":
            dependencies.add(type_info["name"])
        elif type_info["kind"] in ["list", "set"]:
            process_type_info(type_info["item_type"])
        elif type_info["kind"] == "dict":
            process_type_info(type_info["key_type"])
            process_type_info(type_info["value_type"])
        elif type_info["kind"] == "union":
            for t in type_info["types"]:
                process_type_info(t)
        elif type_info["kind"] == "generic":
            for arg in type_info["args"]:
                process_type_info(arg)

    for field_info in model_def["fields"].values():
        process_type_info(field_info["type"])

    return dependencies


def topological_sort(serialized_models: dict[str, dict[str, Any]]) -> list[str]:
    """Sort model names in dependency order."""
    # Build dependency graph
    graph = {
        name: get_model_dependencies(model_def)
        for name, model_def in serialized_models.items()
    }

    # Topological sort using DFS
    visited = set()
    temp = set()
    order = []

    def visit(node: str):
        if node in temp:
            raise ValueError(f"Circular dependency detected involving {node}")
        if node in visited:
            return

        temp.add(node)
        for dep in graph[node]:
            if dep in graph:  # Only process dependencies that are in our models
                visit(dep)
        temp.remove(node)
        visited.add(node)
        order.append(node)

    for node in graph:
        if node not in visited:
            visit(node)

    return order


def deserialize_all_models(serialized_data: dict[str, Any]) -> dict[str, type[BaseModel]]:
    """Deserialize all models from a single dictionary"""
    serialized_models = serialized_data["models"]
    model_registry = {}

    # First pass: create placeholder classes for all models
    for model_name in serialized_models:
        placeholder = type(model_name, (BaseModel,), {})
        model_registry[model_name] = placeholder

    # Sort models by dependency order
    model_order = topological_sort(serialized_models)

    # Second pass: fill in the actual definitions in dependency order
    for model_name in model_order:
        deserialize_model_definition(serialized_models[model_name], model_registry)

    # Return the complete registry of models
    return model_registry


def deserialize_model_definition(
    model_def: dict[str, Any], model_registry: dict[str, type[BaseModel]]
) -> type[BaseModel]:
    """Recreate a Pydantic model class from its serialized definition."""
    model_name = model_def["name"]

    # Process fields
    fields = {}
    for field_name, field_info in model_def["fields"].items():
        # Deserialize field type
        field_type = deserialize_type(field_info["type"], model_registry)

        # Process field parameters
        field_params = {}
        if not field_info["required"]:
            default_value = field_info["default"]

            # Handle default value that is a model
            if isinstance(default_value, dict) and default_value.get("is_model"):
                model_type = field_type
                if hasattr(model_type, "__origin__"):  # Handle Optional[Model] etc.
                    for arg in get_args(model_type):
                        if isinstance(arg, type) and issubclass(arg, BaseModel):
                            model_type = arg
                            break

                if model_type in model_registry:
                    model_class = model_registry[model_type]
                    field_params["default"] = model_class(**default_value["value"])
                else:
                    # Skip default if model not yet available
                    pass
            else:
                field_params["default"] = default_value

        if field_info["description"]:
            field_params["description"] = field_info["description"]

        # Add validation to accept instances of original model classes
        if isinstance(field_type, type) and issubclass(field_type, BaseModel):
            # Ensure the model type is properly registered
            if field_type.__name__ in model_registry:
                field_type = model_registry[field_type.__name__]

            def validate_model(value):
                if isinstance(value, field_type):
                    return value
                if isinstance(value, dict):
                    return field_type(**value)
                raise ValueError(
                    f"Expected {field_type.__name__} or dict, got {type(value)}"
                )

            field_params["validator"] = validate_model

        fields[field_name] = (field_type, Field(**field_params))

    # Create the model with appropriate configuration
    config_dict = model_def.get("config", {})

    # Create actual model class
    if config_dict:
        actual_model = create_model(
            model_name, **fields, __config__=type("Config", (), config_dict)
        )
    else:
        actual_model = create_model(model_name, **fields)

    model_registry[model_name] = actual_model

    return actual_model


# Simplified API functions for serializing and deserializing
def serialize_model(model_class: type[Any]) -> dict[str, Any]:
    """Serialize a model class and all its dependencies to JSON"""
    return serialize_all_models(model_class)


def deserialize_model(model_dict: dict[str, Any]) -> type[Any]:
    """Deserialize model classes from JSON"""
    root_model = model_dict["root_model"]
    model_dict = deserialize_all_models(model_dict)
    return model_dict[root_model]



import typing as t
from typing import Any, Optional

from pydantic import BaseModel, Field

from . import ToolOutput


class CellValue(BaseModel):
    """Model representing a cell's value"""

    value: Any | None = Field(default=None, description="Cell value")
    formula: Optional[str] = Field(default=None, description="Cell formula")
    data_type: str = Field(default="str", description="Cell value type")


class CellData(BaseModel):
    """Model representing a cell's data"""

    value: CellValue = Field(..., description="Cell value")
    row: int = Field(..., description="Row index (1-based)")
    column: int = Field(..., description="Column index (1-based)")


class CellRange(BaseModel):
    """Model representing a range of cells in a sheet"""

    start_row: int = Field(..., description="Start row index (1-based)")
    start_column: int = Field(..., description="Start column index (1-based)")
    end_row: int = Field(..., description="End row index (1-based)")
    end_column: int = Field(..., description="End column index (1-based)")


class CreateSheetInput(BaseModel):
    datasource_id: int = Field(
        default=0, description="ID of the datasource for the Excel file"
    )
    sheet_name: str = Field(description="Name of the new sheet")
    cells: list[CellData] = Field(
        default_factory=list, description="Cells to populate in the new sheet"
    )


class CreateSheetOutput(ToolOutput):
    sheet_name: str = Field(description="Name of the new sheet")


class EditCellOperation(BaseModel):
    """Configuration for editing a cell."""

    sheet_name: str = Field(description="Name of the sheet")
    row: int = Field(description="Row index (1-based)")
    column: int | str = Field(description="Column index or name")
    value: t.Any = Field(description="New value for the cell")


class EditCellInput(BaseModel):
    datasource_id: int = Field(
        default=0, description="ID of the datasource for the Excel file"
    )
    operations: list[EditCellOperation] = Field(default_factory=list)


class EditCellOutput(ToolOutput):
    pass


class UpdateSheetInput(BaseModel):
    datasource_id: int = Field(
        default=0, description="ID of the datasource for the Excel file"
    )

    sheet_name: str = Field(description="Name of the sheet")
    start_row: int = Field(description="Start row index (1-based)")
    start_column: int = Field(description="Start column index (1-based)")
    values: list[list[CellValue]] = Field(description="2D array of values to set")
    batch_size: int = Field(
        default=2, description="Number of columns to update at once (default: 5)"
    )


class UpdateSheetOutput(ToolOutput):
    pass


class ReadSheetInput(BaseModel):
    datasource_id: int = Field(
        default=0, description="ID of the datasource for the Excel file"
    )

    sheet_name: str = Field(description="Name of the sheet to read")
    cell_range: CellRange | None = Field(None, description="Range of cells to read")
    recalculate_formula: bool = Field(
        default=True, description="Whether to recalculate formulas"
    )
    parse_dates_as_epoch: bool = Field(
        default=False, description="Whether to parse Excel dates as epoch timestamps"
    )


class ReadSheetOutput(ToolOutput):
    sheet_name: str = Field(description="Name of the sheet")
    cells: list[CellData] = Field(description="Cells in the sheet")
    range: CellRange | None = Field(None, description="Range of cells")


class ListSheetInput(BaseModel):
    datasource_id: int = Field(
        default=0, description="ID of the datasource for the Excel file"
    )


class ListSheetOutput(ToolOutput):
    sheet_names: list[str] = Field(description="List of sheet names")



import typing as t

from pydantic import BaseModel

from . import ToolOutput


class OutlookInput(BaseModel):
    access_token: str
    folder_name: str = "News Flow"
    start_date: t.Optional[str] = None
    end_date: t.Optional[str] = None


class EmailItem(BaseModel):
    subject: str
    body: str
    date: str
    sender: t.Optional[str] = None
    to: t.Optional[list[str]] = None


class OutlookOutput(ToolOutput):
    content: list[EmailItem]



import typing as t

from pydantic import BaseModel, Field

from . import ToolOutput


class TableInfo(BaseModel):
    """Table information in a Snowflake connection."""

    name: str
    type: str
    id: str
    schemaStripe: str
    databaseStripe: str


class ConnectionDetails(BaseModel):
    """Detailed information about a Snowflake connection."""

    name: str
    type: str
    id: str
    scheduled: bool
    connectionType: str
    configuration: str
    tables: list[TableInfo] = []


class ConnectionResponse(BaseModel):
    """Response model for Snowflake connection creation."""

    id: str
    name: str
    data_warehouse_type: str
    details: ConnectionDetails


class SearchDataResponse(BaseModel):
    """Response model for the search_data API."""

    column_names: list[str]
    available_data_row_count: int


class CreateConnectionInput(BaseModel):
    """Configuration for creating a connection to a data source."""

    data_store_connector_id: str | None = None
    data_store_connector_name: str | None = Field(
        None,  # if None, we will use the name of the Connector from ID.
        description="Name of the connector whose payload will be used to create a "
        "Thoughtspot connection (e.g., Snowflake)",
    )
    data_store_config: dict[str, t.Any] = Field(default_factory=dict)


class CreateConnectionOutput(ToolOutput):
    connection: ConnectionResponse


class CreateLiveboardInput(BaseModel):
    """Configuration for creating a Thoughtspot liveboard."""

    liveboard_name: str = Field(description="Name of the liveboard to create")
    liveboard_description: str | None = Field(
        None, description="Description of the liveboard to create"
    )


class CreateLiveboardOutput(ToolOutput):
    liveboard_id: str


class PinAnswerInput(BaseModel):
    """Configuration for pinning an answer to a liveboard."""

    table_id: str = Field(description="ID of the table to pin an answer to")
    answer_name: str = Field(description="Name of the answer to pin")
    liveboard_id: str = Field(description="ID of the liveboard to pin to")
    chart_type: str = Field(description="Type of chart to use for the answer")
    answer_data: SearchDataResponse = Field(description="Data for the answer")


class PinAnswerOutput(ToolOutput):
    liveboard_id: str = Field(description="ID of the liveboard to pin an answer to")


class SearchMetadataInput(BaseModel):
    """Configuration for searching metadata in Thoughtspot."""

    query: str = Field(description="Query to search for")


class SearchMetadataOutput(ToolOutput):
    result: list[dict[str, t.Any]]


class SearchDataInput(BaseModel):
    """Configuration for searching data in Thoughtspot."""

    query: str = Field(description="Query to search for")
    logical_table_identifier: str = Field(
        description="Logical table identifier to search for"
    )


class SearchDataOutput(ToolOutput):
    result: SearchDataResponse



import os
from dataclasses import dataclass
from enum import Enum


@dataclass(frozen=True)
class ContentTypeInfo:
    mime_type: str
    extensions: list[str]


class ContentType(Enum):
    DOC = ContentTypeInfo(mime_type="application/msword", extensions=["doc"])
    DOCX = ContentTypeInfo(
        mime_type="application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        extensions=["docx"],
    )
    CSV = ContentTypeInfo(mime_type="text/csv", extensions=["csv"])
    PDF = ContentTypeInfo(mime_type="application/pdf", extensions=["pdf"])
    PPTX = ContentTypeInfo(
        mime_type="application/vnd.openxmlformats-officedocument.presentationml.presentation",
        extensions=["pptx"],
    )
    XLSX = ContentTypeInfo(
        mime_type="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        extensions=["xlsx", "xlsm"],
    )
    PPT = ContentTypeInfo(mime_type="application/vnd.ms-powerpoint", extensions=["ppt"])
    HTML = ContentTypeInfo(mime_type="text/html", extensions=["html", "htm"])
    PLAIN_TEXT = ContentTypeInfo(
        mime_type="text/plain", extensions=["md", "markdown", "txt", "text"]
    )
    PNG = ContentTypeInfo(mime_type="image/png", extensions=["png"])
    JPEG = ContentTypeInfo(mime_type="image/jpeg", extensions=["jpg", "jpeg"])
    MP3 = ContentTypeInfo(mime_type="audio/mpeg", extensions=["mp3"])
    MPGA = ContentTypeInfo(mime_type="audio/mpeg", extensions=["mpga"])
    WAV = ContentTypeInfo(mime_type="audio/wav", extensions=["wav"])
    M4A = ContentTypeInfo(mime_type="audio/mp4", extensions=["m4a"])
    OGG = ContentTypeInfo(mime_type="audio/ogg", extensions=["ogg"])
    MPEG = ContentTypeInfo(mime_type="video/mpeg", extensions=["mpeg", "mpg"])
    MP4 = ContentTypeInfo(mime_type="video/mp4", extensions=["mp4"])
    WEBM = ContentTypeInfo(mime_type="video/webm", extensions=["webm"])
    ZIP = ContentTypeInfo(mime_type="application/zip", extensions=["zip"])
    DEFAULT_UPLOAD_TYPE = ContentTypeInfo(
        mime_type="application/octet-stream", extensions=[]
    )
    EML = ContentTypeInfo(mime_type="message/rfc822", extensions=["eml"])

    @classmethod
    def get_all_supported_extensions(cls) -> set[str]:
        """Returns a set of all supported file extensions."""
        extensions = set()
        for content_type in cls:
            extensions.update(content_type.value.extensions)
        return extensions

    @classmethod
    def get_all_mime_types(cls) -> set[str]:
        """Returns a set of all supported file extensions."""
        extensions = set()
        for content_type in cls:
            extensions.add(content_type.value.mime_type)
        return extensions

    @classmethod
    def from_filepath(cls, filepath: str) -> "ContentType":
        _, extension = os.path.splitext(filepath)
        return cls.from_extension(extension)

    @classmethod
    def from_extension(cls, extension: str) -> "ContentType":
        """
        Get ContentType from file extension.
        Raises ValueError if extension is not supported.
        """
        extension = extension.lower().lstrip(".")
        for content_type in cls:
            if extension in content_type.value.extensions:
                return content_type
        raise ValueError(f"Unsupported file extension: {extension}")

    @classmethod
    def from_mime_type(cls, mime_type: str) -> "ContentType":
        """
        Get ContentType from MIME type.
        Raises ValueError if MIME type is not supported.
        """
        for content_type in cls:
            if content_type.value.mime_type == mime_type:
                return content_type
        raise ValueError(f"Unsupported MIME type: {mime_type}")


class RunStatus(str, Enum):
    # NOTE: When new statuses are added, alembic does not pick up the new enum
    # values. So, we need to create an alembic migration to include the
    # new statuses.
    UNDEFINED = "undefined"
    NOT_STARTED = "not_started"
    PENDING = "pending"
    RUNNING = "running"
    FAILED = "failed"
    SUCCEEDED = "succeeded"
    PARTIAL_SUCCESS = "partial_success"
    SKIPPED = "skipped"
    PAUSED = "paused"
    CANCELLED = "cancelled"


class CustomDataType(str, Enum):
    TEXT = "text"
    BOOL = "bool"
    EMAIL = "email"
    DATETIME = "datetime"
    TIME = "time"
    DATE = "date"
    NUMBER = "number"
    DOCUMENT = "document"
    FOLDER = "folder"
    INPUT_TEMPLATE = "input_template"
    URL = "url"
    EXCEL = "excel"
    ENUM = "enum"


class ModelName(str, Enum):
    """
    Enum class for standardized model names used across the codebase.
    """

    GEMINI_FLASH = "gemini/gemini-2.5-flash-preview-05-20"
    GEMINI_PRO = "gemini/gemini-2.5-pro-preview-05-06"
    GPT_4O = "gpt-4o"
    CLAUDE_SONNET = "claude-3-5-sonnet-20240620"
    GPT_4O_MINI = "gpt-4o-mini"
    GEMINI_FLASH_1_5 = "gemini/gemini-1.5-flash"
    O4_MINI = "o4-mini"
    O3 = "o3"
